#include "GPstation.h"

#define CPU_IPS			0x10000
#define GPU_IPS			0x12000

SDL_Surface *sdl_screen;
bool req_reflesh = false;

void load_prg(gp3200_cpu *core, int baseaddr, char *binname, int maxsize);

/*
	-- Memory Map (CPU) --
	0x00000000-0x0000ffff : IVT(64KB)
	0x00010000-0x0030ffff : VRAM(32bpp,3MB)
	0x00310000-0x0130ffff : GIOS(16MB)
	0x01310000-0x2130ffff : GAME PRG(512MB)
	0x21310000-0x2131ffff : Stack
	0x21320000-0x40000000 : Free Area
	Total : 1GB
	
	GPU Memory : 256MB
	0x00000000-0x002fffff : VRAM(CPU Memory link)
	0x00300000-0x012fffff : GIOS for GPU(16MB)
	0x01300000-0x0130ffff : Stack
	0x01310000-0x10000000 : Free Area
*/

int main(int argc, char **argv) {
	SDL_Event sdl_event;
	gp3200_cpu core_gp3200_cpu;
	gp3200_cpu core_gp3200_gpu;
	UINT8 *cpu_memory;
	UINT8 *gpu_memory;
	int cpu_ips;
	int gpu_ips;

	/*if(argc < 2) {
		puts("using>GPstation <romfile>\n");
		exit(-1);
	}*/

	cpu_memory = (UINT8 *)malloc(1024*1024*1024);
	gpu_memory = (UINT8 *)malloc(256*1024*1024);

	if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO|SDL_INIT_TIMER)) exit(1);

	sdl_screen = SDL_SetVideoMode(640,480,32,SDL_HWSURFACE);
	SDL_WM_SetCaption("GPstation emu", NULL);

	gp3200_init(&core_gp3200_cpu, cpu_memory);
	gp3200_init(&core_gp3200_gpu, gpu_memory);

	core_gp3200_cpu.r_pc = 0x00310000;
	core_gp3200_gpu.r_pc = 0x00300000;
	core_gp3200_cpu.r_sp = 0x21320000;
	core_gp3200_gpu.r_sp = 0x01310000;
	core_gp3200_cpu.r_segs[0] = 0;
	core_gp3200_gpu.r_segs[0] = 0;

	core_gp3200_gpu.r_memory[0x00300000] = 0xff;

	memset(core_gp3200_cpu.r_port, 0, 256 * sizeof(UINT32));

	int i;
	for(i = 0; i < 32; i++) core_gp3200_cpu.r_regs[i] = 0;
	for(i = 0; i < 32; i++) core_gp3200_gpu.r_regs[i] = 0;

	load_prg(&core_gp3200_cpu,0x00310000,"gios.bin",16*1024*1024);

	while(!poll_event(&sdl_event)) {
		for(cpu_ips = 0; cpu_ips < CPU_IPS; cpu_ips++) gp3200_step(&core_gp3200_cpu);
		for(gpu_ips = 0; gpu_ips < GPU_IPS; gpu_ips++) gp3200_step(&core_gp3200_gpu);

		reflesh_io(&core_gp3200_cpu);
		//link_vram(cpu_memory, gpu_memory);
		if(req_reflesh == true) {
			reflesh_vram(cpu_memory);
			SDL_UpdateRect(sdl_screen,0,0,0,0);
			req_reflesh = false;
		}
	}
}

void load_prg(gp3200_cpu *core, int baseaddr, char *binname, int maxsize)
{
	FILE *fp = fopen(binname,"rb");
	if(fp == NULL) return;

	fread(core->r_memory + baseaddr,sizeof(char),maxsize,fp);

	fclose(fp);
}