/*
	GP3200 EMU maincode
	cording by taka
	version : 2.1
*/

#include "gp3200.h"
#include <stdlib.h>

void gp3200_get_int1(UINT32 **data, gp3200_cpu *cpu)
{
	UINT8 opcode = cpu->r_memory[cpu->r_pc + 0];
	UINT8 type = cpu->r_memory[cpu->r_pc + 1] & 0xf;
	UINT32 args1 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 2]);

	cpu->r_pc += 2;
	
	switch(type) {
	case ARGS_IR:
		cpu->r_pc += 1;
		*data = &cpu->r_regs[args1 & 0xff];
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		*data = (UINT32 *)&cpu->r_memory[cpu->r_regs[args1 & 0xff] + cpu->r_segs[1]];
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		*data = (UINT32 *)&cpu->r_memory[args1 + cpu->r_segs[1]];
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data = args1;
		break;
	case ARGS_SR:
		cpu->r_pc += 1;
		*data = &cpu->r_segs[args1 & 0xff];
		break;
	}
};
	
void gp3200_get_int2(UINT32 **data1, UINT32 **data2, gp3200_cpu *cpu)
{
	UINT8 opcode = cpu->r_memory[cpu->r_pc + 0];
	UINT8 type[2];
	UINT32 args1 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 2]);
	UINT32 args2;
	
	type[0] = cpu->r_memory[cpu->r_pc + 1] & 0xf;
	type[1] = (cpu->r_memory[cpu->r_pc + 1] >> 4) & 0xf;

	cpu->r_pc += 2;
	
	switch(type[0]) {
	case ARGS_IR:
		cpu->r_pc += 1;
		*data1 = &cpu->r_regs[args1 & 0xff];
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		*data1 = (UINT32 *)&cpu->r_memory[cpu->r_regs[args1 & 0xff] + cpu->r_segs[1]];
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		*data1 = (UINT32 *)&cpu->r_memory[args1 + cpu->r_segs[1]];
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data1 = args1;
		break;
	case ARGS_SR:
		cpu->r_pc += 1;
		*data1 = &cpu->r_segs[args1 & 0xff];
		break;
	}
	
	args2 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 0]);
	
	switch(type[1]) {
	case ARGS_IR:
		cpu->r_pc += 1;
		*data2 = &cpu->r_regs[args2 & 0xff];
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		*data2 = (UINT32 *)&cpu->r_memory[cpu->r_regs[args2 & 0xff] + cpu->r_segs[1]];
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		*data2 = (UINT32 *)&cpu->r_memory[args2 + cpu->r_segs[1]];
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data2 = args2;
		break;
	case ARGS_FR:
		cpu->r_pc += 1;
		**data2 = (UINT32)gp3200_f32_getfloat(cpu->f_regs[args2 & 0xff]);
		if(type[0] == ARGS_MR || type[0] == ARGS_ME) *data2 = (UINT32 *)&cpu->f_regs[args2 & 0xff];
		break;
	case ARGS_FE:
		cpu->r_pc += 4;
		**data2 = (UINT32)gp3200_f32_getfloat(gp3200_f32_int2float(args2));
		break;
	case ARGS_SR:
		cpu->r_pc += 1;
		*data2 = &cpu->r_segs[args2 & 0xff];
		break;
	}
};

void gp3200_get_float2(gp3200_float **data1, gp3200_float **data2, gp3200_cpu *cpu)
{
	UINT8 opcode = cpu->r_memory[cpu->r_pc + 0];
	UINT8 type[2];
	UINT32 args1 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 2]);
	UINT32 args2;
	
	type[0] = cpu->r_memory[cpu->r_pc + 1] & 0xf;
	type[1] = (cpu->r_memory[cpu->r_pc + 1] >> 4) & 0xf;

	cpu->r_pc += 2;
	
	switch(type[0]) {
	case ARGS_FR:
		cpu->r_pc += 1;
		*data1 = &cpu->f_regs[args1 & 0xff];
		break;
	}
	
	args2 = *((UINT32 *)&cpu->r_memory[cpu->r_pc + 0]);
	
	switch(type[1]) {
	case ARGS_FR:
		cpu->r_pc += 1;
		*data2 = &cpu->f_regs[args2 & 0xff];
		break;
	case ARGS_IR:
		cpu->r_pc += 1;
		**data2 = gp3200_f32_int2float(cpu->r_regs[args2 & 0xff]);
		break;
	case ARGS_MR:
		cpu->r_pc += 1;
		**data2 = gp3200_f32_int2float(cpu->r_memory[cpu->r_regs[args2 & 0xff] + cpu->r_segs[1]]);
		break;
	case ARGS_ME:
		cpu->r_pc += 4;
		**data2 = gp3200_f32_int2float(cpu->r_memory[args2 + cpu->r_segs[1]]);
		break;
	case ARGS_E:
		cpu->r_pc += 4;
		**data2 = gp3200_f32_int2float(args2);
		break;
	case ARGS_FE:
		cpu->r_pc += 4;
		**data2 = *((gp3200_float *)&args2);
		break;
	}
};

void gp3200_init(gp3200_cpu *cpu, UINT8 *memory)
{
	int i;
	cpu->r_port = (UINT32 *)malloc(256 * sizeof(UINT32));
	cpu->r_if = 1;
	for(i = 0; i < 8; i++) cpu->r_segs[i] = 0;
	cpu->r_memory = memory;
	cpu->on_int = 0;
}

void gp3200_step(gp3200_cpu *cpu)
{
	unsigned int used, used1, used2;
	unsigned int *data = &used, *data1 = &used1, *data2 = &used2;
	UINT8 opcode = cpu->r_memory[(cpu->r_pc & (0x80000000 - 1)) + cpu->r_segs[0]];
	
	switch(opcode) {
	case 0x00:		// NOP
		cpu->r_pc++;
		break;
	case 0x01:		// MOV
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 = *data2;
		break;
	case 0x02:		// ADD
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 += *data2;
		break;
	case 0x03:		// SUB
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 -= *data2;
		break;
	case 0x04:		// IMUL
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 *= *data2;
		break;
	case 0x05:		// IDIV
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 /= *data2;
		break;
	case 0x06:		// AND
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 &= *data2;
		break;
	case 0x07:		// OR
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 |= *data2;
		break;
	case 0x08:		// NOT
		gp3200_get_int1(&data, cpu);
		*data = ~*data;
		break;
	case 0x09:		// XOR
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 ^= *data2;
		break;
	case 0x0a:		// INT
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_if == 1) {
			if(cpu->on_int == 0) {
				cpu->r_pc = *((UINT32 *)&cpu->r_memory[*data * 8]);
				cpu->r_segs[0] = *((UINT32 *)&cpu->r_memory[*data * 8 + 4]);
			}
			else cpu->on_int(cpu, *data, cpu->r_memory);
		}
		break;
	case 0x0b:		// CALL
		gp3200_get_int1(&data, cpu);
		cpu->r_ret = cpu->r_pc;
		cpu->r_pc = *data;
		break;
	case 0x0c:		// RET
		cpu->r_pc = cpu->r_ret;
		break;
	case 0x0d:		// JMP
		gp3200_get_int1(&data, cpu);
		cpu->r_pc = *data;
		break;
	case 0x0e:		// CMP
		gp3200_get_int2(&data1, &data2, cpu);
		
		if(*data1 == *data2     ) cpu->r_cmp |= 1 << 0;
		if(*data1 >= *data2     ) cpu->r_cmp |= 1 << 1;
		if(*data1 <= *data2     ) cpu->r_cmp |= 1 << 2;
		if(*data1 >  *data2     ) cpu->r_cmp |= 1 << 3;
		if(*data1 <  *data2     ) cpu->r_cmp |= 1 << 4;
		if(*data1 -  *data2 == 0) cpu->r_cmp |= 1 << 5;
		if(*data1 -  *data2 != 0) cpu->r_cmp |= 1 << 6;
		if(*data1 != *data2     ) cpu->r_cmp |= 1 << 7;
		break;
	case 0x0f:		// SAR
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 >>= *data2;
		break;
	case 0x10:		// SAL
		gp3200_get_int2(&data1, &data2, cpu);
		*data1 <<= *data2;
		break;
	case 0x11:		// JE
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 0))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x12:		// JAE
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 1))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x13:		// JBE
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 2))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x14:		// JA
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 3))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x15:		// JB
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 4))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x16:		// JZ
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 5))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x17:		// JNZ
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 6))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x18:		// JNE
		gp3200_get_int1(&data, cpu);
		
		if(cpu->r_cmp & (1 << 7))
		{
			cpu->r_pc = *data;
			cpu->r_cmp = 0;
		}
		break;
	case 0x19:		// IN
		gp3200_get_int2(&data1, &data2, cpu);
		*data2 = cpu->r_port[*data1];
		break;
	case 0x1a:		// OUT
		gp3200_get_int2(&data1, &data2, cpu);
		cpu->r_port[*data1] = *data2;
		break;
	case 0x1b:		// PUSH
		gp3200_get_int1(&data, cpu);
		cpu->r_sp -= 4;
		cpu->r_memory[cpu->r_regs[6]] = *data;
		break;
	case 0x1c:		// POP
		gp3200_get_int1(&data, cpu);
		cpu->r_sp += 4;
		*data = cpu->r_memory[cpu->r_regs[6]];
		break;
	case 0x1d:		// CLI
		cpu->r_if = 0;
		cpu->r_pc++;
		break;
	case 0x1e:		// STI
		cpu->r_if = 1;
		cpu->r_pc++;
		break;
	case 0x1f:		// FMOV
		gp3200_get_float2(&data1, &data2, cpu);
		*data1 = *data2;
		break;
	case 0x20:		// FADD
		gp3200_get_float2(&data1, &data2, cpu);
		*data1 = gp3200_f32_fadd(*data1, *data2);
		break;
	case 0x21:		// FSUB
		gp3200_get_float2(&data1, &data2, cpu);
		*data1 = gp3200_f32_fsub(*data1, *data2);
		break;
	case 0x22:		// FMUL
		gp3200_get_float2(&data1, &data2, cpu);
		*data1 = gp3200_f32_fmul(*data1, *data2);
		break;
	case 0x23:		// FDIV
		gp3200_get_float2(&data1, &data2, cpu);
		*data1 = gp3200_f32_fdiv(*data1, *data2);
		break;
/*  --- v2 or higher opcode --- */
	case 0x24:		// FARJMP
		gp3200_get_int2(&data1, &data2, cpu);
		cpu->r_pc = *data1;
		cpu->r_segs[0] = *data2;
		break;
	case 0x25:		// FARCALL
		gp3200_get_int2(&data1, &data2, cpu);
		cpu->r_ret = cpu->r_pc;
		cpu->r_retcs = cpu->r_segs[0];
		cpu->r_pc = *data1;
		cpu->r_segs[0] = *data2;
		break;
	case 0x26:		// FARRET
		cpu->r_pc = cpu->r_ret;
		cpu->r_segs[0] = cpu->r_retcs;
		break;
	case 0x27:		// LSP
		gp3200_get_int1(&data, cpu);
		cpu->r_sp = *data;
		break;
	case 0x28:		// SSP
		gp3200_get_int1(&data, cpu);
		*data = cpu->r_sp;
		break;
	default:
		break;
	}
}
